module Borges

##
# Grab the Session for this thread and ...

def respond
  Thread.current[:Session].respond { |r| yield r }
end

##
# A Session tracks the state of an application

class Session

  attr :last_access
  attr :session_mutex
  attr :thread_mutex
  attr :thread

  attr_reader :continuations
  
  ##
  # Create a new session for app

  def initialize(app)
    @session_mutex = Mutex.new
    @thread_mutex = Mutex.new.lock
    @application = app
    @continuations = LRUCache.new

    @thread = Thread.start do
      begin
        cur_thr = Thread.current
        cur_thr[:Session] = self
        cur_thr.abort_on_exception = true

        loop do
          @last_access = Time.now
          @thread_mutex.synchronize do
            handle_request_intern(cur_thr[:request], cur_thr[:response])
            cur_thr[:response_done].unlock
          end
          @thread_mutex.lock
        end
      rescue Exception => err
        err_str = %!Exception in #{cur_thr.inspect}:\n#{err}\n#{err.backtrace.join("\n")}!
        cur_thr[:response].error(err_str, cur_thr[:request])
        cur_thr[:response_done].unlock
        cur_thr.kill
      end
    end
  end

  ##
  # Set up the Session's thread for execution, and dispatch

  def handle_request(request, response)
    @session_mutex.synchronize do
      response_lock = Mutex.new.lock
      @thread[:request] = request
      @thread[:response] = response
      @thread[:response_done] = response_lock
      @thread_mutex.unlock
      response_lock.lock
    end
  end
  
  ##
  # Take the continuation and save it in the cache. ...

  def respond
    callcc do |cc|
      @response.action_key = @continuations.store(cc)
      yield(@response)
      @short_jmp.call
    end
  end

  ##
  # Kill this thread
  # TODO: look into finalizers

  def expire
    @session_mutex.synchronize { @thread.kill }
  end

  private
 
  ##
  # Handle a request, dispatching to the request's action_key, if such a key
  # exists in the hash, otherwise start the application from the beginning.
  # 
  # Called by the public handle_request

  def handle_request_intern(request, response)
    @response = response
    @current_key = request.action_key
    callcc do |cc|
      @short_jmp = cc
      if cont = @continuations[@current_key]
        # if there is a continuation for current_key, invoke
        cont.call(request)
      else
        # handle unknown requests
        unknown_request(request)
      end
    end
    @last_key = @current_key
  end

  ##
  # Start this request over

  def unknown_request(request)
    @application.entry_point.call
  end

  ##
  # Spit out a backtrace for the Exception err

  def handle_error(err)
    @response.body = "<h1>#{err.message}</h1><pre>#{err.backtrace.join('\n')}</pre>"
  end
end

end

